# Stubs for unittest

# Based on http://docs.python.org/2.7/library/unittest.html

# Only a subset of functionality is included.

from typing import (
    Any, Callable, Iterable, Tuple, List, TextIO, Sequence,
    overload, TypeVar, Pattern
)
from abc import abstractmethod, ABCMeta

_T = TypeVar('_T')
_FT = TypeVar('_FT')

class Testable(metaclass=ABCMeta):
    @abstractmethod
    def run(self, result: 'TestResult') -> None: ...
    @abstractmethod
    def debug(self) -> None: ...
    @abstractmethod
    def countTestCases(self) -> int: ...

# TODO ABC for test runners?

class TestResult:
    errors = ... # type: List[Tuple[Testable, str]]
    failures = ... # type: List[Tuple[Testable, str]]
    testsRun = 0
    shouldStop = False

    def wasSuccessful(self) -> bool: ...
    def stop(self) -> None: ...
    def startTest(self, test: Testable) -> None: ...
    def stopTest(self, test: Testable) -> None: ...
    def addError(self, test: Testable,
                  err: Tuple[type, Any, Any]) -> None: ... # TODO
    def addFailure(self, test: Testable,
                    err: Tuple[type, Any, Any]) -> None: ... # TODO
    def addSuccess(self, test: Testable) -> None: ...

class _AssertRaisesBaseContext:
    expected = ... # type: Any
    failureException = ... # type: type
    obj_name = ...  # type: str
    expected_regex = ... # type: Pattern[str]

class _AssertRaisesContext(_AssertRaisesBaseContext):
    exception = ... # type: Any # TODO precise type
    def __enter__(self) -> _AssertRaisesContext: ...
    def __exit__(self, exc_type, exc_value, tb) -> bool: ...

class _AssertWarnsContext(_AssertRaisesBaseContext):
    warning = ... # type: Any # TODO precise type
    filename = ...  # type: str
    lineno = 0
    def __enter__(self) -> _AssertWarnsContext: ...
    def __exit__(self, exc_type, exc_value, tb) -> bool: ...

class TestCase(Testable):
    def __init__(self, methodName: str = ...) -> None: ...
    # TODO failureException
    def setUp(self) -> None: ...
    def tearDown(self) -> None: ...
    def run(self, result: TestResult = ...) -> None: ...
    def debug(self) -> None: ...
    def assert_(self, expr: Any, msg: object = ...) -> None: ...
    def failUnless(self, expr: Any, msg: object = ...) -> None: ...
    def assertTrue(self, expr: Any, msg: object = ...) -> None: ...
    def assertEqual(self, first: Any, second: Any,
                    msg: object = ...) -> None: ...
    def failUnlessEqual(self, first: Any, second: Any,
                        msg: object = ...) -> None: ...
    def assertNotEqual(self, first: Any, second: Any,
                       msg: object = ...) -> None: ...
    def assertSequenceEqual(self, first: Sequence[Any], second: Sequence[Any],
                            msg: object = ...,
                            seq_type: type = ...) -> None: ...
    def failIfEqual(self, first: Any, second: Any,
                    msg: object = ...) -> None: ...
    def assertAlmostEqual(self, first: float, second: float, places: int = ...,
                          msg: object = ...,
                          delta: float = ...) -> None: ...
    def failUnlessAlmostEqual(self, first: float, second: float,
                              places: int = ...,
                              msg: object = ...) -> None: ...
    def assertNotAlmostEqual(self, first: float, second: float,
                             places: int = ..., msg: object = ...,
                             delta: float = ...) -> None: ...
    def failIfAlmostEqual(self, first: float, second: float, places: int = ...,
                          msg: object = ...) -> None: ...
    def assertGreater(self, first: Any, second: Any,
                      msg: object = ...) -> None: ...
    def assertGreaterEqual(self, first: Any, second: Any,
                      msg: object = ...) -> None: ...
    def assertLess(self, first: Any, second: Any,
                   msg: object = ...) -> None: ...
    def assertLessEqual(self, first: Any, second: Any,
                        msg: object = ...) -> None: ...
    def assertRaises(self, expected_exception: type, *args: Any, **kwargs: Any) -> Any: ...
    def failIf(self, expr: Any, msg: object = ...) -> None: ...
    def assertFalse(self, expr: Any, msg: object = ...) -> None: ...
    def assertIs(self, first: object, second: object,
                 msg: object = ...) -> None: ...
    def assertIsNot(self, first: object, second: object,
                    msg: object = ...) -> None: ...
    def assertIsNone(self, expr: Any, msg: object = ...) -> None: ...
    def assertIsNotNone(self, expr: Any, msg: object = ...) -> None: ...
    def assertIn(self, first: _T, second: Iterable[_T],
                 msg: object = ...) -> None: ...
    def assertNotIn(self, first: _T, second: Iterable[_T],
                    msg: object = ...) -> None: ...
    def assertIsInstance(self, obj: Any, cls: type,
                         msg: object = ...) -> None: ...
    def assertNotIsInstance(self, obj: Any, cls: type,
                            msg: object = ...) -> None: ...
    def assertWarns(self, expected_warning: type, callable_obj: Any = ...,
                    *args: Any, **kwargs: Any) -> _AssertWarnsContext: ...
    def fail(self, msg: object = ...) -> None: ...
    def countTestCases(self) -> int: ...
    def defaultTestResult(self) -> TestResult: ...
    def id(self) -> str: ...
    def shortDescription(self) -> str: ... # May return None
    def addCleanup(function: Any, *args: Any, **kwargs: Any) -> None: ...
    def skipTest(self, reason: Any) -> None: ...

    assertEquals = assertEqual
    assertNotEquals = assertNotEqual
    assertAlmostEquals = assertAlmostEqual
    assertNotAlmostEquals = assertNotAlmostEqual
    assert_ = assertTrue

    failUnlessEqual = assertEqual
    failIfEqual = assertNotEqual
    failUnlessAlmostEqual = assertAlmostEqual
    failIfAlmostEqual = assertNotAlmostEqual
    failUnless = assertTrue
    failUnlessRaises = assertRaises
    failIf = assertFalse

class CallableTestCase(Testable):
    def __init__(self, testFunc: Callable[[], None],
                 setUp: Callable[[], None] = ...,
                 tearDown: Callable[[], None] = ...,
                 description: str = ...) -> None: ...
    def run(self, result: TestResult) -> None: ...
    def debug(self) -> None: ...
    def countTestCases(self) -> int: ...

class TestSuite(Testable):
    def __init__(self, tests: Iterable[Testable] = ...) -> None: ...
    def addTest(self, test: Testable) -> None: ...
    def addTests(self, tests: Iterable[Testable]) -> None: ...
    def run(self, result: TestResult) -> None: ...
    def debug(self) -> None: ...
    def countTestCases(self) -> int: ...

# TODO TestLoader
# TODO defaultTestLoader

class TextTestRunner:
    def __init__(self, stream: TextIO = ..., descriptions: bool = ...,
                 verbosity: int = ..., failfast: bool = ...) -> None: ...

class SkipTest(Exception):
    ...

# TODO precise types
def skipUnless(condition: Any, reason: str) -> Any: ...
def skipIf(condition: Any, reason: str) -> Any: ...
def expectedFailure(func: _FT) -> _FT: ...
def skip(reason: str) -> Any: ...

def main(module: str = ..., defaultTest: str = ...,
         argv: List[str] = ..., testRunner: Any = ...,
         testLoader: Any = ...) -> None: ... # TODO types



# Stubs for gzip (Python 2)
#
# NOTE: Based on a dynamically typed stub automatically generated by stubgen.

from typing import Any, IO
import io

class GzipFile(io.BufferedIOBase):
    myfileobj = ... # type: Any
    max_read_chunk = ... # type: Any
    mode = ... # type: Any
    extrabuf = ... # type: Any
    extrasize = ... # type: Any
    extrastart = ... # type: Any
    name = ... # type: Any
    min_readsize = ... # type: Any
    compress = ... # type: Any
    fileobj = ... # type: Any
    offset = ... # type: Any
    mtime = ... # type: Any
    def __init__(self, filename: str = ..., mode: str = ..., compresslevel: int = ...,
                 fileobj: IO[str] = ..., mtime: float = ...) -> None: ...
    @property
    def filename(self): ...
    size = ... # type: Any
    crc = ... # type: Any
    def write(self, data): ...
    def read(self, size=...): ...
    @property
    def closed(self): ...
    def close(self): ...
    def flush(self, zlib_mode=...): ...
    def fileno(self): ...
    def rewind(self): ...
    def readable(self): ...
    def writable(self): ...
    def seekable(self): ...
    def seek(self, offset, whence=...): ...
    def readline(self, size=...): ...

def open(filename: str, mode: str = ..., compresslevel: int = ...) -> GzipFile: ...



import logging
from typing import Any

rootlogger = ... # type: Any

class Identified(object):
    def _should_log_debug(self) -> bool: ...
    def _should_log_info(self) -> bool: ...

class InstanceLogger(object): ...

def instance_logger(instance, echoflag) -> None: ...

class echo_property(object): ...



# Stubs for Crypto.Cipher (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

# Names in __all__ with no definition:
#   AES
#   ARC2
#   ARC4
#   Blowfish
#   CAST
#   DES
#   DES3
#   PKCS1_OAEP
#   PKCS1_v1_5
#   XOR



# Generated by stubtool 0.1, DO NOT EDIT
# See https://github.com/o11c/stubtool
#
# Stubs for six.moves.urllib.request (Python 3.2)

from urllib.request import BaseHandler as BaseHandler
from urllib.request import HTTPRedirectHandler as HTTPRedirectHandler
from urllib.request import OpenerDirector as OpenerDirector

from urllib.request import install_opener as install_opener
from urllib.request import build_opener as build_opener

# from urllib.request import urlopen as urlopen
# from urllib.request import pathname2url as pathname2url
# from urllib.request import url2pathname as url2pathname
# from urllib.request import getproxies as getproxies
# from urllib.request import Request as Request
# from urllib.request import HTTPDefaultErrorHandler as HTTPDefaultErrorHandler
# from urllib.request import HTTPCookieProcessor as HTTPCookieProcessor
# from urllib.request import ProxyHandler as ProxyHandler
# from urllib.request import HTTPPasswordMgr as HTTPPasswordMgr
# from urllib.request import HTTPPasswordMgrWithDefaultRealm as HTTPPasswordMgrWithDefaultRealm
# from urllib.request import AbstractBasicAuthHandler as AbstractBasicAuthHandler
# from urllib.request import HTTPBasicAuthHandler as HTTPBasicAuthHandler
# from urllib.request import ProxyBasicAuthHandler as ProxyBasicAuthHandler
# from urllib.request import AbstractDigestAuthHandler as AbstractDigestAuthHandler
# from urllib.request import HTTPDigestAuthHandler as HTTPDigestAuthHandler
# from urllib.request import ProxyDigestAuthHandler as ProxyDigestAuthHandler
# from urllib.request import HTTPHandler as HTTPHandler
# from urllib.request import HTTPSHandler as HTTPSHandler
# from urllib.request import FileHandler as FileHandler
# from urllib.request import FTPHandler as FTPHandler
# from urllib.request import CacheFTPHandler as CacheFTPHandler
# from urllib.request import UnknownHandler as UnknownHandler
# from urllib.request import HTTPErrorProcessor as HTTPErrorProcessor
# from urllib.request import urlretrieve as urlretrieve
# from urllib.request import urlcleanup as urlcleanup
# from urllib.request import URLopener as URLopener
# from urllib.request import FancyURLopener as FancyURLopener
from urllib.request import proxy_bypass as proxy_bypass



from typing import Any
# Stubs for abc.

# Thesee definitions have special processing in type checker.
class ABCMeta:
    def register(cls: "ABCMeta", subclass: Any) -> None: ...
abstractmethod = object()
abstractproperty = object()



# Stubs for sqlalchemy.util (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from . import compat
from . import _collections
from . import langhelpers
from . import deprecations

callable = compat.callable
cmp = compat.cmp
reduce = compat.reduce
threading = compat.threading
py3k = compat.py3k
py33 = compat.py33
py2k = compat.py2k
jython = compat.jython
pypy = compat.pypy
cpython = compat.cpython
win32 = compat.win32
pickle = compat.pickle
dottedgetter = compat.dottedgetter
parse_qsl = compat.parse_qsl
namedtuple = compat.namedtuple
next = compat.next
reraise = compat.reraise
raise_from_cause = compat.raise_from_cause
text_type = compat.text_type
safe_kwarg = compat.safe_kwarg
string_types = compat.string_types
int_types = compat.int_types
binary_type = compat.binary_type
nested = compat.nested
quote_plus = compat.quote_plus
with_metaclass = compat.with_metaclass
print_ = compat.print_
itertools_filterfalse = compat.itertools_filterfalse
u = compat.u
ue = compat.ue
b = compat.b
unquote_plus = compat.unquote_plus
unquote = compat.unquote
b64decode = compat.b64decode
b64encode = compat.b64encode
byte_buffer = compat.byte_buffer
itertools_filter = compat.itertools_filter
iterbytes = compat.iterbytes
StringIO = compat.StringIO
inspect_getargspec = compat.inspect_getargspec
zip_longest = compat.zip_longest
KeyedTuple = _collections.KeyedTuple
ImmutableContainer = _collections.ImmutableContainer
immutabledict = _collections.immutabledict
Properties = _collections.Properties
OrderedProperties = _collections.OrderedProperties
ImmutableProperties = _collections.ImmutableProperties
OrderedDict = _collections.OrderedDict
OrderedSet = _collections.OrderedSet
IdentitySet = _collections.IdentitySet
OrderedIdentitySet = _collections.OrderedIdentitySet
column_set = _collections.column_set
column_dict = _collections.column_dict
ordered_column_set = _collections.ordered_column_set
populate_column_dict = _collections.populate_column_dict
unique_list = _collections.unique_list
UniqueAppender = _collections.UniqueAppender
PopulateDict = _collections.PopulateDict
EMPTY_SET = _collections.EMPTY_SET
to_list = _collections.to_list
to_set = _collections.to_set
to_column_set = _collections.to_column_set
update_copy = _collections.update_copy
flatten_iterator = _collections.flatten_iterator
has_intersection = _collections.has_intersection
LRUCache = _collections.LRUCache
ScopedRegistry = _collections.ScopedRegistry
ThreadLocalRegistry = _collections.ThreadLocalRegistry
WeakSequence = _collections.WeakSequence
coerce_generator_arg = _collections.coerce_generator_arg
lightweight_named_tuple = _collections.lightweight_named_tuple
iterate_attributes = langhelpers.iterate_attributes
class_hierarchy = langhelpers.class_hierarchy
portable_instancemethod = langhelpers.portable_instancemethod
unbound_method_to_callable = langhelpers.unbound_method_to_callable
getargspec_init = langhelpers.getargspec_init
format_argspec_init = langhelpers.format_argspec_init
format_argspec_plus = langhelpers.format_argspec_plus
get_func_kwargs = langhelpers.get_func_kwargs
get_cls_kwargs = langhelpers.get_cls_kwargs
decorator = langhelpers.decorator
as_interface = langhelpers.as_interface
memoized_property = langhelpers.memoized_property
memoized_instancemethod = langhelpers.memoized_instancemethod
md5_hex = langhelpers.md5_hex
group_expirable_memoized_property = langhelpers.group_expirable_memoized_property
dependencies = langhelpers.dependencies
decode_slice = langhelpers.decode_slice
monkeypatch_proxied_specials = langhelpers.monkeypatch_proxied_specials
asbool = langhelpers.asbool
bool_or_str = langhelpers.bool_or_str
coerce_kw_type = langhelpers.coerce_kw_type
duck_type_collection = langhelpers.duck_type_collection
assert_arg_type = langhelpers.assert_arg_type
symbol = langhelpers.symbol
dictlike_iteritems = langhelpers.dictlike_iteritems
classproperty = langhelpers.classproperty
set_creation_order = langhelpers.set_creation_order
warn_exception = langhelpers.warn_exception
warn = langhelpers.warn
NoneType = langhelpers.NoneType
constructor_copy = langhelpers.constructor_copy
methods_equivalent = langhelpers.methods_equivalent
chop_traceback = langhelpers.chop_traceback
asint = langhelpers.asint
generic_repr = langhelpers.generic_repr
counter = langhelpers.counter
PluginLoader = langhelpers.PluginLoader
hybridproperty = langhelpers.hybridproperty
hybridmethod = langhelpers.hybridmethod
safe_reraise = langhelpers.safe_reraise
get_callable_argspec = langhelpers.get_callable_argspec
only_once = langhelpers.only_once
attrsetter = langhelpers.attrsetter
ellipses_string = langhelpers.ellipses_string
warn_limited = langhelpers.warn_limited
map_bits = langhelpers.map_bits
MemoizedSlots = langhelpers.MemoizedSlots
EnsureKWArgType = langhelpers.EnsureKWArgType
warn_deprecated = deprecations.warn_deprecated
warn_pending_deprecation = deprecations.warn_pending_deprecation
deprecated = deprecations.deprecated
pending_deprecation = deprecations.pending_deprecation
inject_docstring_text = deprecations.inject_docstring_text



# Stubs for email.utils (Python 3.4)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

import email._parseaddr

mktime_tz = email._parseaddr.mktime_tz
parsedate = email._parseaddr.parsedate
parsedate_tz = email._parseaddr.parsedate_tz

def formataddr(pair, charset=...): ...
def getaddresses(fieldvalues): ...
def formatdate(timeval=..., localtime=..., usegmt=...): ...
def format_datetime(dt, usegmt=...): ...
def make_msgid(idstring=..., domain=...): ...
def parsedate_to_datetime(data): ...
def parseaddr(addr): ...
def unquote(str): ...
def decode_rfc2231(s): ...
def encode_rfc2231(s, charset=..., language=...): ...
def decode_params(params): ...
def collapse_rfc2231_value(value, errors=..., fallback_charset=...): ...



from typing import Any, IO, Optional, Tuple, Callable, Dict, List

class JSONDecodeError(object):
    def dumps(self, obj: Any) -> str: ...
    def dump(self, obj: Any, fp: IO[str], *args: Any, **kwds: Any) -> None: ...
    def loads(self, s: str) -> Any: ...
    def load(self, fp: IO[str]) -> Any: ...

def dumps(obj: Any,
    skipkeys: bool = ...,
    ensure_ascii: bool = ...,
    check_circular: bool = ...,
    allow_nan: bool = ...,
    cls: Any = ...,
    indent: Optional[int] = ...,
    separators: Optional[Tuple[str, str]] = ...,
    encoding: str = ...,
    default: Optional[Callable[[Any], Any]] = ...,
    sort_keys: bool = ...,
    **kwds: Any) -> str: ...

def dump(obj: Any,
    fp: IO[str],
    skipkeys: bool = ...,
    ensure_ascii: bool = ...,
    check_circular: bool = ...,
    allow_nan: bool = ...,
    cls: Any = ...,
    indent: Optional[int] = ...,
    separators: Optional[Tuple[str, str]] = ...,
    encoding: str = ...,
    default: Optional[Callable[[Any], Any]] = ...,
    sort_keys: bool = ...,
    **kwds: Any) -> None: ...

def loads(s: str,
    encoding: Any = ...,
    cls: Any = ...,
    object_hook: Callable[[Dict], Any] = ...,
    parse_float: Optional[Callable[[str], Any]] = ...,
    parse_int: Optional[Callable[[str], Any]] = ...,
    parse_constant: Optional[Callable[[str], Any]] = ...,
    object_pairs_hook: Optional[Callable[[List[Tuple[Any, Any]]], Any]] = ...,
    **kwds: Any) -> Any: ...

def load(fp: IO[str],
    encoding: Optional[str] = ...,
    cls: Any = ...,
    object_hook: Callable[[Dict], Any] = ...,
    parse_float: Optional[Callable[[str], Any]] = ...,
    parse_int: Optional[Callable[[str], Any]] = ...,
    parse_constant: Optional[Callable[[str], Any]] = ...,
    object_pairs_hook: Optional[Callable[[List[Tuple[Any, Any]]], Any]] = ...,
    **kwds: Any) -> Any: ...



"""Classes for producing HTML reports about imprecision."""

from abc import ABCMeta, abstractmethod
import cgi
import os
import shutil
import tokenize

from typing import Callable, Dict, List, Tuple, cast

from mypy.nodes import MypyFile, Node, FuncDef
from mypy import stats
from mypy.traverser import TraverserVisitor
from mypy.types import Type


reporter_classes = {}  # type: Dict[str, Callable[[Reports, str], AbstractReporter]]


class Reports:
    def __init__(self, data_dir: str, report_dirs: Dict[str, str]) -> None:
        self.data_dir = data_dir
        self.reporters = []  # type: List[AbstractReporter]
        self.named_reporters = {}  # type: Dict[str, AbstractReporter]

        for report_type, report_dir in sorted(report_dirs.items()):
            self.add_report(report_type, report_dir)

    def add_report(self, report_type: str, report_dir: str) -> 'AbstractReporter':
        try:
            return self.named_reporters[report_type]
        except KeyError:
            pass
        reporter_cls = reporter_classes[report_type]
        reporter = reporter_cls(self, report_dir)
        self.reporters.append(reporter)
        self.named_reporters[report_type] = reporter
        return reporter

    def file(self, tree: MypyFile, type_map: Dict[Node, Type]) -> None:
        for reporter in self.reporters:
            reporter.on_file(tree, type_map)

    def finish(self) -> None:
        for reporter in self.reporters:
            reporter.on_finish()


class AbstractReporter(metaclass=ABCMeta):
    def __init__(self, reports: Reports, output_dir: str) -> None:
        self.output_dir = output_dir

    @abstractmethod
    def on_file(self, tree: MypyFile, type_map: Dict[Node, Type]) -> None:
        pass

    @abstractmethod
    def on_finish(self) -> None:
        pass


class FuncCounterVisitor(TraverserVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.counts = [0, 0]

    def visit_func_def(self, defn: FuncDef):
        self.counts[defn.type is not None] += 1


class LineCountReporter(AbstractReporter):
    def __init__(self, reports: Reports, output_dir: str) -> None:
        super().__init__(reports, output_dir)
        self.counts = {}  # type: Dict[str, Tuple[int, int, int, int]]

        stats.ensure_dir_exists(output_dir)

    def on_file(self, tree: MypyFile, type_map: Dict[Node, Type]) -> None:
        physical_lines = len(open(tree.path).readlines())

        func_counter = FuncCounterVisitor()
        tree.accept(func_counter)
        unannotated_funcs, annotated_funcs = func_counter.counts
        total_funcs = annotated_funcs + unannotated_funcs

        imputed_annotated_lines = (physical_lines * annotated_funcs // total_funcs
                                   if total_funcs else physical_lines)

        self.counts[tree._fullname] = (imputed_annotated_lines, physical_lines,
                                       annotated_funcs, total_funcs)

    def on_finish(self) -> None:
        counts = sorted(((c, p) for p, c in self.counts.items()),
                        reverse=True)  # type: List[Tuple[tuple, str]]
        total_counts = tuple(sum(c[i] for c, p in counts)
                             for i in range(4))
        with open(os.path.join(self.output_dir, 'linecount.txt'), 'w') as f:
            f.write('{:7} {:7} {:6} {:6} total\n'.format(*total_counts))
            for c, p in counts:
                f.write('{:7} {:7} {:6} {:6} {}\n'.format(
                    c[0], c[1], c[2], c[3], p))

reporter_classes['linecount'] = LineCountReporter


class OldHtmlReporter(AbstractReporter):
    """Old HTML reporter.

    This just calls the old functions in `stats`, which use global
    variables to preserve state for the index.
    """

    def on_file(self, tree: MypyFile, type_map: Dict[Node, Type]) -> None:
        stats.generate_html_report(tree, tree.path, type_map, self.output_dir)

    def on_finish(self) -> None:
        stats.generate_html_index(self.output_dir)
reporter_classes['old-html'] = OldHtmlReporter


class FileInfo:
    def __init__(self, name: str, module: str) -> None:
        self.name = name
        self.module = module
        self.counts = [0] * len(stats.precision_names)

    def total(self) -> int:
        return sum(self.counts)

    def attrib(self) -> Dict[str, str]:
        return {name: str(val) for name, val in zip(stats.precision_names, self.counts)}


class MemoryXmlReporter(AbstractReporter):
    """Internal reporter that generates XML in memory.

    This is used by all other XML-based reporters to avoid duplication.
    """

    def __init__(self, reports: Reports, output_dir: str) -> None:
        import lxml.etree as etree

        super().__init__(reports, output_dir)

        self.xslt_html_path = os.path.join(reports.data_dir, 'xml', 'mypy-html.xslt')
        self.xslt_txt_path = os.path.join(reports.data_dir, 'xml', 'mypy-txt.xslt')
        self.css_html_path = os.path.join(reports.data_dir, 'xml', 'mypy-html.css')
        xsd_path = os.path.join(reports.data_dir, 'xml', 'mypy.xsd')
        self.schema = etree.XMLSchema(etree.parse(xsd_path))
        self.last_xml = None  # type: etree._ElementTree
        self.files = []  # type: List[FileInfo]

    def on_file(self, tree: MypyFile, type_map: Dict[Node, Type]) -> None:
        import lxml.etree as etree

        self.last_xml = None
        path = os.path.relpath(tree.path)
        if stats.is_special_module(path):
            return
        if path.startswith('..'):
            return
        if 'stubs' in path.split('/'):
            return

        visitor = stats.StatisticsVisitor(inferred=True, typemap=type_map, all_nodes=True)
        tree.accept(visitor)

        root = etree.Element('mypy-report-file', name=path, module=tree._fullname)
        doc = etree.ElementTree(root)
        file_info = FileInfo(path, tree._fullname)

        with tokenize.open(path) as input_file:
            for lineno, line_text in enumerate(input_file, 1):
                status = visitor.line_map.get(lineno, stats.TYPE_EMPTY)
                file_info.counts[status] += 1
                etree.SubElement(root, 'line',
                                 number=str(lineno),
                                 precision=stats.precision_names[status],
                                 content=line_text[:-1])
        # Assumes a layout similar to what XmlReporter uses.
        xslt_path = os.path.relpath('mypy-html.xslt', path)
        transform_pi = etree.ProcessingInstruction('xml-stylesheet',
                'type="text/xsl" href="%s"' % cgi.escape(xslt_path, True))
        root.addprevious(transform_pi)
        self.schema.assertValid(doc)

        self.last_xml = doc
        self.files.append(file_info)

    def on_finish(self) -> None:
        import lxml.etree as etree

        self.last_xml = None
        # index_path = os.path.join(self.output_dir, 'index.xml')
        output_files = sorted(self.files, key=lambda x: x.module)

        root = etree.Element('mypy-report-index', name='index')
        doc = etree.ElementTree(root)

        for file_info in output_files:
            etree.SubElement(root, 'file',
                             file_info.attrib(),
                             total=str(file_info.total()),
                             name=file_info.name,
                             module=file_info.module)
        xslt_path = os.path.relpath('mypy-html.xslt', '.')
        transform_pi = etree.ProcessingInstruction('xml-stylesheet',
                'type="text/xsl" href="%s"' % cgi.escape(xslt_path, True))
        root.addprevious(transform_pi)
        self.schema.assertValid(doc)

        self.last_xml = doc

reporter_classes['memory-xml'] = MemoryXmlReporter


class AbstractXmlReporter(AbstractReporter):
    """Internal abstract class for reporters that work via XML."""

    def __init__(self, reports: Reports, output_dir: str) -> None:
        super().__init__(reports, output_dir)

        memory_reporter = reports.add_report('memory-xml', '<memory>')
        # The dependency will be called first.
        self.memory_xml = cast(MemoryXmlReporter, memory_reporter)


class XmlReporter(AbstractXmlReporter):
    """Public reporter that exports XML.

    The produced XML files contain a reference to the absolute path
    of the html transform, so they will be locally viewable in a browser.

    However, there is a bug in Chrome and all other WebKit-based browsers
    that makes it fail from file:// URLs but work on http:// URLs.
    """

    def on_file(self, tree: MypyFile, type_map: Dict[Node, Type]) -> None:
        last_xml = self.memory_xml.last_xml
        if last_xml is None:
            return
        path = os.path.relpath(tree.path)
        if path.startswith('..'):
            return
        out_path = os.path.join(self.output_dir, 'xml', path + '.xml')
        stats.ensure_dir_exists(os.path.dirname(out_path))
        last_xml.write(out_path, encoding='utf-8')

    def on_finish(self) -> None:
        last_xml = self.memory_xml.last_xml
        out_path = os.path.join(self.output_dir, 'index.xml')
        out_xslt = os.path.join(self.output_dir, 'mypy-html.xslt')
        out_css = os.path.join(self.output_dir, 'mypy-html.css')
        last_xml.write(out_path, encoding='utf-8')
        shutil.copyfile(self.memory_xml.xslt_html_path, out_xslt)
        shutil.copyfile(self.memory_xml.css_html_path, out_css)
        print('Generated XML report:', os.path.abspath(out_path))

reporter_classes['xml'] = XmlReporter


class XsltHtmlReporter(AbstractXmlReporter):
    """Public reporter that exports HTML via XSLT.

    This is slightly different than running `xsltproc` on the .xml files,
    because it passes a parameter to rewrite the links.
    """

    def __init__(self, reports: Reports, output_dir: str) -> None:
        import lxml.etree as etree

        super().__init__(reports, output_dir)

        self.xslt_html = etree.XSLT(etree.parse(self.memory_xml.xslt_html_path))
        self.param_html = etree.XSLT.strparam('html')

    def on_file(self, tree: MypyFile, type_map: Dict[Node, Type]) -> None:
        last_xml = self.memory_xml.last_xml
        if last_xml is None:
            return
        path = os.path.relpath(tree.path)
        if path.startswith('..'):
            return
        out_path = os.path.join(self.output_dir, 'html', path + '.html')
        stats.ensure_dir_exists(os.path.dirname(out_path))
        transformed_html = bytes(self.xslt_html(last_xml, ext=self.param_html))
        with open(out_path, 'wb') as out_file:
            out_file.write(transformed_html)

    def on_finish(self) -> None:
        last_xml = self.memory_xml.last_xml
        out_path = os.path.join(self.output_dir, 'index.html')
        out_css = os.path.join(self.output_dir, 'mypy-html.css')
        transformed_html = bytes(self.xslt_html(last_xml, ext=self.param_html))
        with open(out_path, 'wb') as out_file:
            out_file.write(transformed_html)
        shutil.copyfile(self.memory_xml.css_html_path, out_css)
        print('Generated HTML report (via XSLT):', os.path.abspath(out_path))

reporter_classes['xslt-html'] = XsltHtmlReporter


class XsltTxtReporter(AbstractXmlReporter):
    """Public reporter that exports TXT via XSLT.

    Currently this only does the summary, not the individual reports.
    """

    def __init__(self, reports: Reports, output_dir: str) -> None:
        import lxml.etree as etree

        super().__init__(reports, output_dir)

        self.xslt_txt = etree.XSLT(etree.parse(self.memory_xml.xslt_txt_path))

    def on_file(self, tree: MypyFile, type_map: Dict[Node, Type]) -> None:
        pass

    def on_finish(self) -> None:
        last_xml = self.memory_xml.last_xml
        out_path = os.path.join(self.output_dir, 'index.txt')
        stats.ensure_dir_exists(os.path.dirname(out_path))
        transformed_txt = bytes(self.xslt_txt(last_xml))
        with open(out_path, 'wb') as out_file:
            out_file.write(transformed_txt)
        print('Generated TXT report (via XSLT):', os.path.abspath(out_path))

reporter_classes['xslt-txt'] = XsltTxtReporter

reporter_classes['html'] = reporter_classes['xslt-html']
reporter_classes['txt'] = reporter_classes['xslt-txt']



# Stubs for requests.api (Python 3)

import typing

from .models import Response

def request(method: str, url: str, **kwargs) -> Response: ...
def get(url: str, **kwargs) -> Response: ...
def options(url: str, **kwargs) -> Response: ...
def head(url: str, **kwargs) -> Response: ...
def post(url: str, data=..., json=..., **kwargs) -> Response: ...
def put(url: str, data=..., **kwargs) -> Response: ...
def patch(url: str, data=..., **kwargs) -> Response: ...
def delete(url: str, **kwargs) -> Response: ...



# Stubs for types
# Note, all classes "defined" here require special handling.

# TODO parts of this should be conditional on version

from typing import Any, Callable, Dict, Iterator, Optional, Tuple, TypeVar, Union, overload

_T = TypeVar('_T')

class _Cell:
    cell_contents = ...  # type: Any

class FunctionType:
    __closure__ = ...  # type: Optional[Tuple[_Cell, ...]]
    __code__ = ...  # type: CodeType
    __defaults__ = ...  # type: Optional[Tuple[Any, ...]]
    __dict__ = ...  # type: Dict[str, Any]
    __doc__ = ...  # type: Optional[str]
    __globals__ = ...  # type: Dict[str, Any]
    __name__ = ...  # type: str
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
    def __get__(self, obj: Optional[object], type: Optional[type]) -> 'MethodType': ...
LambdaType = FunctionType

class CodeType:
    """Create a code object.  Not for the faint of heart."""
    co_argcount = ... # type: int
    co_kwonlyargcount = ... # type: int
    co_nlocals = ... # type: int
    co_stacksize = ... # type: int
    co_flags = ... # type: int
    co_code = ... # type: bytes
    co_consts = ... # type: Tuple[Any, ...]
    co_names = ... # type: Tuple[str, ...]
    co_varnames = ... # type: Tuple[str, ...]
    co_filename = ... # type: Optional[str]
    co_name = ... # type: str
    co_firstlineno = ... # type: int
    co_lnotab = ... # type: bytes
    co_freevars = ... # type: Tuple[str, ...]
    co_cellvars = ... # type: Tuple[str, ...]
    def __init__(self,
            argcount: int,
            kwonlyargcount: int,
            nlocals: int,
            stacksize: int,
            flags: int,
            codestring: bytes,
            constants: Tuple[Any, ...],
            names: Tuple[str, ...],
            varnames: Tuple[str, ...],
            filename: str,
            name: str,
            firstlineno: int,
            lnotab: bytes,
            freevars: Tuple[str, ...] = ...,
            cellvars: Tuple[str, ...] = ...,
    ) -> None: ...

class MappingProxyType:
    def copy(self) -> dict: ...
    def get(self, key: str, default: _T = ...) -> Union[Any, _T]: ...
    def items(self) -> Iterator[Tuple[str, Any]]: ...
    def keys(self) -> Iterator[str]: ...
    def values(self) -> Iterator[Any]: ...
    def __contains__(self, key: str) -> bool: ...
    def __getitem__(self, key: str) -> Any: ...
    def __iter__(self) -> Iterator[str]: ...
    def __len__(self) -> int: ...
class SimpleNamespace(Any): ...

class GeneratorType:
    gi_code = ...  # type: CodeType
    gi_frame = ...  # type: FrameType
    gi_running = ...  # type: bool
    gi_yieldfrom = ...  # type: Optional[GeneratorType]
    def __iter__(self) -> 'GeneratorType': ...
    def __next__(self) -> Any: ...
    def close(self) -> None: ...
    def send(self, arg: Any) -> Any: ...
    @overload
    def throw(self, val: BaseException) -> Any: ...
    @overload
    def throw(self, typ: type, val: BaseException = ..., tb: 'TracebackType' = ...) -> Any: ...

class CoroutineType:
    cr_await = ...  # type: Optional[Any]
    cr_code = ...  # type: CodeType
    cr_frame = ...  # type: FrameType
    cr_running = ...  # type: bool
    def close(self) -> None: ...
    def send(self, arg: Any) -> Any: ...
    @overload
    def throw(self, val: BaseException) -> Any: ...
    @overload
    def throw(self, typ: type, val: BaseException = ..., tb: 'TracebackType' = ...) -> Any: ...

class MethodType:
    __func__ = ...  # type: FunctionType
    __self__ = ...  # type: object
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
class BuiltinFunctionType:
    __self__ = ...  # type: Union[object, ModuleType]
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
BuiltinMethodType = BuiltinFunctionType

class ModuleType:
    __name__ = ... # type: str
    __file__ = ... # type: str
    def __init__(self, name: str, doc: Any) -> None: ...

class TracebackType:
    tb_frame = ... # type: FrameType
    tb_lasti = ... # type: int
    tb_lineno = ... # type: int
    tb_next = ... # type: TracebackType

class FrameType:
    f_back = ... # type: FrameType
    f_builtins = ... # type: Dict[str, Any]
    f_code = ... # type: CodeType
    f_globals = ... # type: Dict[str, Any]
    f_lasti = ... # type: int
    f_lineno = ... # type: int
    f_locals = ... # type: Dict[str, Any]
    f_trace = ... # type: Callable[[], None]

    def clear(self) -> None: pass

class GetSetDescriptorType:
    __name__ = ...  # type: str
    __objclass__ = ...  # type: type
    def __get__(self, obj: Any, type: type = ...) -> Any: ...
    def __set__(self, obj: Any) -> None: ...
    def __delete__(self, obj: Any) -> None: ...
class MemberDescriptorType:
    __name__ = ...  # type: str
    __objclass__ = ...  # type: type
    def __get__(self, obj: Any, type: type = ...) -> Any: ...
    def __set__(self, obj: Any) -> None: ...
    def __delete__(self, obj: Any) -> None: ...

def new_class(name: str, bases: Tuple[type, ...] = ..., kwds: Dict[str, Any] = ..., exec_body: Callable[[Dict[str, Any]], None] = ...) -> type: ...
def prepare_class(name: str, bases: Tuple[type, ...] = ..., kwds: Dict[str, Any] = ...) -> Tuple[type, Dict[str, Any], Dict[str, Any]]: ...

# Actually a different type, but `property` is special and we want that too.
DynamicClassAttribute = property

def coroutine(f: Callable[..., Any]) -> CoroutineType: ...



from typing import Sequence, TypeVar

_T = TypeVar('_T')

def insort_left(a: Sequence[_T], x: _T, lo: int = ..., hi: int = ...): pass
def insort_right(a: Sequence[_T], x: _T, lo: int = ..., hi: int = ...): pass

def bisect_left(a: Sequence[_T], x: _T, lo: int = ..., hi: int = ...): pass
def bisect_right(a: Sequence[_T], x: _T, lo: int = ..., hi: int = ...): pass

insort = insort_right
bisect = bisect_right



# Stubs for struct

# Based on http://docs.python.org/3.2/library/struct.html

from typing import overload, Any, AnyStr, Tuple

class error(Exception): ...

def pack(fmt: AnyStr, *v: Any) -> bytes: ...
# TODO buffer type
def pack_into(fmt: AnyStr, buffer: Any, offset: int, *v: Any) -> None: ...

# TODO buffer type
def unpack(fmt: AnyStr, buffer: Any) -> Tuple[Any, ...]: ...
def unpack_from(fmt: AnyStr, buffer: Any, offset: int = ...) -> Tuple[Any, ...]: ...

def calcsize(fmt: AnyStr) -> int: ...

class Struct:
    format = b''
    size = 0

    def __init__(self, format: AnyStr) -> None: ...

    def pack(self, *v: Any) -> bytes: ...
    # TODO buffer type
    def pack_into(self, buffer: Any, offset: int, *v: Any) -> None: ...
    # TODO buffer type
    def unpack(self, buffer: Any) -> Tuple[Any, ...]: ...
    def unpack_from(self, buffer: Any, offset: int = ...) -> Tuple[Any, ...]: ...



from typing import Any, List

default_action = ...  # type: str
filters = ...  # type: List[tuple]
once_registry = ...  # type: dict

def warn(message: Warning, category:type = ..., stacklevel:int = ...) -> None: ...
def warn_explicit(message: Warning, category:type,
                  filename: str, lineno: int,
                  module:Any = ..., registry:dict = ...,
                  module_globals:dict = ...) -> None: ...
